-- settings ------------------------------------------------------------------

-- we use eldritch-world.cpk because it contains only rooms and thus is easily checkable
GAME_ROOT = "../../.."
WORLD_CPK = GAME_ROOT .. "/eldritch-world.cpk"
STREAM_FILE = "" -- set during pkg tests

-- utility functions ---------------------------------------------------------
function ANNOUNCE(test)
	printf("Running %s", test)
end

function CHECK(test)
	if not test then
		local info = debug.getinfo(2, "l")
		printf("  CHECK FAILED in line %d", info.currentline)
		return false
	end
	return true
end

function FLTEQ(a, b)
	return math.abs(a-b) < math.min(a, b) / 10000
end

-- tests start here
local TESTS = {}

-- test for pkg library ------------------------------------------------------
function test_pkg()
	ANNOUNCE("test_pkg")

	local ok = CHECK(pkg.add("doesnotexist.cpk") == false)
	
	ok = ok and CHECK(pkg.add(WORLD_CPK))

	local tbl = pkg.list(WORLD_CPK)
	ok = ok and CHECK(tbl and #tbl > 0)

	for i = 1, #tbl do
		local file = tbl[i]
		ok = ok and CHECK(string.find(file, "Rooms/.*%.eldritchroom") ~= nil)
		ok = ok and CHECK(pkg.fileexists(file))
	end

	STREAM_FILE = tbl[1] -- side effect, just to later have a file that exists

	ok = ok and CHECK(pkg.fileexists("doesnotexist.eldritchroom") == false)

	return ok
end
TESTS[#TESTS+1] = test_pkg

-- test for config library and GETVAR handler --------------------------------
do
	local old_log = log
	local LOG = {}

	local function test_log(msg, ...)
		local str = string.format(msg, ...)
		LOG[#LOG + 1] = str
	end

	function test_config()
		ANNOUNCE("test_config")

		log = test_log
		local ok = CHECK(config.load("test.config"))
		log = old_log

		ok = ok and CHECK(LOG[1] == "Hello from test.config!")
		ok = ok and CHECK(LOG[2] == "Hello from Context2")
		ok = ok and CHECK(LOG[3] == "Hello from function test in Context1")

		local val

		val = config.getvar("Var1", "")
		ok = ok and CHECK(math.type(val)=="integer" and val==1)

		val = config.getvar("Var2", "")
		ok = ok and CHECK(math.type(val)=="float" and FLTEQ(val, 2.0))

		val = config.getvar("Var3", "")
		ok = ok and CHECK(type(val)=="boolean" and val==true)

		val = config.getvar("Var4", "")
		ok = ok and CHECK(type(val)=="string" and val=="Test")

		val = config.getvar("Var1", "Context1")
		ok = ok and CHECK(math.type(val)=="integer" and val==11)

		val = config.getvar("Var2", "Context1")
		ok = ok and CHECK(math.type(val)=="float" and FLTEQ(val, 2.2))

		val = config.getvar("Var3", "Context1")
		ok = ok and CHECK(type(val)=="boolean" and val==true)

		val = config.getvar("Var4", "Context1")
		ok = ok and CHECK(type(val)=="string" and val=="Test Context1")

		val = config.getvar("Var1", "Context2")
		ok = ok and CHECK(math.type(val)=="integer" and val==111)

		val = config.getvar("Var2", "Context2")
		ok = ok and CHECK(math.type(val)=="float" and FLTEQ(val, 22.2))

		val = config.getvar("Var3", "Context2")
		ok = ok and CHECK(type(val)=="boolean" and val==true)

		val = config.getvar("Var4", "Context2")
		ok = ok and CHECK(type(val)=="string" and val=="Test Context2")

		val = config.getvar("Var1", "Context3")
		ok = ok and CHECK(math.type(val)=="integer" and val==1111)

		val = config.getvar("Var2", "Context3")
		ok = ok and CHECK(math.type(val)=="float" and FLTEQ(val, 222.2))

		val = config.getvar("Var3", "Context3")
		ok = ok and CHECK(type(val)=="boolean" and val==false)

		val = config.getvar("Var4", "Context3")
		ok = ok and CHECK(type(val)=="string" and val=="Some Value")

		val = config.getvar("Var1", "Context4")
		ok = ok and CHECK(type(val)=="string" and val==string.format("[Context4] %08X:%08X", config.hash("Var1"), config.hash("Context4")))

		val = config.getvar("Var2", "Context4")
		ok = ok and CHECK(type(val)=="string" and val==string.format("[Context4] %08X:%08X", config.hash("Var2"), config.hash("Context4")))

		val = config.getvar("Var3", "Context4")
		ok = ok and CHECK(type(val)=="string" and val==string.format("[] %08X:%08X", config.hash("Var3"), config.hash("Context4")))

		val = config.getvar("Var4", "Context4")
		ok = ok and CHECK(type(val)=="string" and val=="a real variable")

		ok = ok and CHECK(config.reversehash(config.hash("Context4")) == "Context4")

		config.registerhash("ImaginaryContext5")
		ok = ok and CHECK(config.reversehash(config.hash("ImaginaryContext5")) == "ImaginaryContext5")

		return ok
	end
end
TESTS[#TESTS+1] = test_config

-- tests for stream library --------------------------------------------------
do
	function test_stream()
		ANNOUNCE("test_stream")

		local ok = CHECK(stream.open("doesnotexist.doesnotexist") == nil)
		local strm = stream.open(STREAM_FILE)
		ok = ok and CHECK(strm ~= nil)
		local ssize = strm:size()
		ok = ok and CHECK(ssize > 0)
		local data = strm:read(ssize + 1)
		ok = ok and CHECK(data == nil)
		ok = ok and CHECK(strm:getpos() == 0)
		ok = ok and CHECK(strm:ateof() == false)
		data = strm:read(ssize)
		ok = ok and CHECK(data ~= nil)
		ok = ok and CHECK(strm:getpos() == ssize)
		strm:setpos()
		ok = ok and CHECK(strm:getpos() == 0)
		local data1 = strm:read()
		ok = ok and CHECK(data == data1)
		ok = ok and CHECK(strm:ateof())

		local lstrm = stream.create()
		lstrm:write(data)
		ok = ok and CHECK(lstrm:size() == ssize)
		ok = ok and CHECK(lstrm:getpos() == ssize)
		lstrm:setpos()
		ok = ok and CHECK(lstrm:getpos() == 0)
		ok = ok and CHECK(lstrm:ateof() == false)
		data1 = lstrm:read(ssize + 1)
		ok = ok and CHECK(data1 == nil)
		ok = ok and CHECK(lstrm:getpos() == 0)
		data1 = lstrm:read()
		ok = ok and CHECK(data == data1)
		ok = ok and CHECK(lstrm:ateof())

		return ok
	end
end
TESTS[#TESTS+1] = test_stream


-- tests for fs library ------------------------------------------------------
function test_fs()
	ANNOUNCE("test_fs")

	-- path normalizing is tricky business...
	local ok = CHECK(fs.normalize("a") == "a")
	ok = ok and CHECK(fs.normalize("a/") == "a")
	ok = ok and CHECK(fs.normalize("a///") == "a")
	ok = ok and CHECK(fs.normalize("/a") == "/a")
	ok = ok and CHECK(fs.normalize("///a") == "/a")
	ok = ok and CHECK(fs.normalize("a/b") == "a/b")
	ok = ok and CHECK(fs.normalize("a\\b") == "a/b")
	ok = ok and CHECK(fs.normalize("a/b///") == "a/b")
	ok = ok and CHECK(fs.normalize("a///b") == "a/b")
	ok = ok and CHECK(fs.normalize("a/\\/b") == "a/b")
	ok = ok and CHECK(fs.normalize("a/b/c") == "a/b/c")
	ok = ok and CHECK(fs.normalize("/a/.") == "/a")
	ok = ok and CHECK(fs.normalize("/a/./././") == "/a")
	ok = ok and CHECK(fs.normalize("/a/./././b") == "/a/b")
	ok = ok and CHECK(fs.normalize("a/./././b") == "a/b")
	ok = ok and CHECK(fs.normalize("a/./b/././c") == "a/b/c")
	ok = ok and CHECK(fs.normalize(".") == ".")
	ok = ok and CHECK(fs.normalize("./a") == "./a")
	ok = ok and CHECK(fs.normalize("././a") == "./a")
	ok = ok and CHECK(fs.normalize("a/..") == "")
	ok = ok and CHECK(fs.normalize("a/../b") == "b")
	ok = ok and CHECK(fs.normalize("a/b/c/../../d") == "a/d")
	ok = ok and CHECK(fs.normalize("..") == "..")
	ok = ok and CHECK(fs.normalize("../..") == "../..")
	ok = ok and CHECK(fs.normalize("../../..") == "../../..")
	ok = ok and CHECK(fs.normalize("../a/..") == "..")
	ok = ok and CHECK(fs.normalize("../a/../..") == "../..")
	ok = ok and CHECK(fs.normalize("a/../../") == "..")
	ok = ok and CHECK(fs.normalize("a/.././../b/..///.///..//") == "../..")
	ok = ok and CHECK(fs.normalize("/../a/b/c") == "")

	local CWD = fs.getcurrentdir()
	ok = ok and CHECK(CWD and #CWD > 0)
	ok = ok and CHECK(fs.isdir(CWD))
	local P1 =  CWD .. "/" .. ARG[0]
	ok = ok and CHECK(fs.isfile(P1))
	ok = ok and CHECK(fs.dirname(P1) == CWD)
	ok = ok and CHECK(fs.filename(P1) == ARG[0])
	ok = ok and CHECK(fs.extension(P1) == "lua")
	ok = ok and CHECK(fs.basename(P1) == fs.basename(ARG[0]))
	ok = ok and CHECK(fs.getexecdir() == fs.normalize(fs.getcurrentdir() .. "/../.."))
	ok = ok and CHECK(fs.getexecname() == "mEldritchLua" or fs.getexecname() == "mEldritchLua.exe")
	ok = ok and CHECK(fs.getmoddir() == fs.getcurrentdir())

	fs.setcurrentdir("..")
	ok = ok and CHECK(fs.getcurrentdir() == fs.normalize(CWD .. "/.."))
	fs.setcurrentdir(CWD)

	local dirlist = fs.list(GAME_ROOT)
	ok = ok and CHECK(dirlist and #dirlist > 0)

	for _, entry in ipairs(dirlist) do
		ok = ok and CHECK(fs.filename(entry) == entry)
	end

	return ok
end
TESTS[#TESTS+1] = test_fs

-- test for texture library --------------------------------------------------
do
	function make_texture()
		local txt = texture.create(16, 16)
		for x=0, 15 do
			for y=0, 15 do
				local grey = x * y + 1
				txt:setpixel(x, y, grey, grey, grey, 255)
			end
		end
		local strm = stream.create()
		txt:saveto(strm)
		return strm
	end

	local function compare_textures(tx1, tx2)
		local sx, sy = tx1.sizex, tx1.sizey
		if sx~=tx2.sizex or sy~=tx2.sizey then
			return false
		end
		for x=0, sx-1 do
			for y=0, sy-1 do
				local r1, g1, b1, a1 = tx1:getpixel(x, y)
				local r2, g2, b2, a2 = tx2:getpixel(x, y)
				if r1~=r2 or g1~=g2 or b1~=b2 or a1~=a1 then
					return false
				end
			end
		end
		return true
	end

	function test_texture()
		ANNOUNCE("test_texture")

		local TXNAME = "test.png"

		local ok = CHECK(texture.load("doesnotexist.png") == nil)
		local tx1 = texture.load("&make_texture() --.png") 
		ok = ok and CHECK(tx1 ~= nil)
		tx1:save(TXNAME)
		local tx2 = texture.load(TXNAME)
		ok = ok and CHECK(tx2 ~= nil)
		ok = ok and CHECK(compare_textures(tx1, tx2))
		local tx3 = texture.create(tx2.sizex, tx2.sizey)
		tx2:copyrect(0, 0, tx2.sizex, tx2.sizey, tx3, 0, 0) -- TODO more tests for copyrect
		ok = ok and CHECK(compare_textures(tx2, tx3))

		os.remove(TXNAME)
		return ok
	end
end
TESTS[#TESTS+1] = test_texture

-- test for room library -----------------------------------------------------
do
	function make_room(worlddef)
		local rm = room.create()
		for x=0, rm.sizex-1 do
			for y=0, rm.sizey-1 do
				rm:setvoxel(x, y, 0, 1)
				rm:setvoxel(x, y, rm.sizez-1, 1)
			end
		end
		for x=0, rm.sizex-1 do
			for z=0, rm.sizez-1 do
				rm:setvoxel(x, 0, z, 1)
				rm:setvoxel(x, rm.sizey-1, z, 1)
			end
		end
		for y=0, rm.sizey-1 do
			for z=0, rm.sizez-1 do
				rm:setvoxel(0, y, z, 1)
				rm:setvoxel(rm.sizex-1, y, z, 1)
			end
		end
		rm:setworlddef(worlddef)
		local strm = stream.create()
		rm:saveto(strm)
		return strm
	end

	local function compare_rooms(rm1, rm2)
		local sx,  sy, sz = rm1.sizex, rm1.sizey, rm1.sizez
		if sx~=rm2.sizex or sy~=rm2.sizey or sz~=rm2.sizez then
			return false
		end
		if rm1:getworlddef()~=rm2:getworlddef() then
			return false
		end
		for x=0, sx-1 do
			for y=0, sy-1 do
				for z=0, sz-1 do
					v1 = rm1:getvoxel(x, y, z)
					v2 = rm2:getvoxel(x, y, z)
					if v1~=v2 then
						return false
					end
				end
			end
		end
		return true
	end

	function test_room()
		ANNOUNCE("test_room")

		local RMNAME = 'test-x.eldritchroom'

		local ok = CHECK(room.load("doesnotexist.eldritchroom") == nil)
		local rm1 = room.load("&make_room('" .. RMNAME .. "', 'Test')")
		ok = ok and CHECK(rm1 ~= nil)
		rm1:save(RMNAME)
		local rm2 = room.load(RMNAME)
		ok = ok and CHECK(rm2 ~= nil)
		ok = ok and  CHECK(compare_rooms(rm1,  rm2))

		local rm3 = room:create()
		rm3:setspawner(1, 2, 3, "One", "n")
		rm3:setspawner(4, 5, 6, "Two", "s")
		rm3:setspawner(7, 6, 5, "Three", "w")
		rm3:setspawner(4, 3, 2, "Four", "e")
		local i = 0
		repeat
			local n1, o1, x1, y1, z1 = rm3:getspawner(i)
			if not n1 then
				break
			end
			local n2, o2, x2, y2, z2 = rm3:getspawner(x1, y1, z1)
			ok = ok and CHECK(n1==n2 and o1==o2 and x1==x2 and y1==y2 and z1==z2)
			i = i + 1
		until false

		os.remove(RMNAME)
		return ok
	end
end
TESTS[#TESTS+1] = test_room

-- run all tests -------------------------------------------------------------
local nfailed = 0
for test=1, #TESTS do
	if TESTS[test]() then
		print("-> OK")
	else
		nfailed = nfailed + 1
		print("-> FAILED")
	end
end
printf("Tests total: %d\nTests failed: %d", #TESTS, nfailed)