# XML Mesh exporter for Blender 2.6+
# port of xmlmeshexport.py
# Gunnar Zötl <gz@tset.de> 2015

bl_info = {
    "name":         "Eldritch XML Mesh (.mesh)",
    "author":       "Gunnar Zötl",
    "blender":      (2,6,2),
    "version":      (0,0,1),
    "location":     "File > Import-Export",
    "description":  "Export selected mesh to Eldritch Engine XML Mesh Format (.mesh)",
    "category":     "Import-Export"
}

import bpy;
from bpy_extras.io_utils import ExportHelper
from mathutils import Matrix

class EldritchXMLMeshExporter(bpy.types.Operator, ExportHelper):
    bl_idname       = "eldritch_xml_mesh.mesh";
    bl_label        = "Export Eldritch XML Mesh";
    bl_options      = {'PRESET'};

    filename_ext    = ".mesh";

    def execute(self, context):
        # Ensure Blender is currently in OBJECT mode to allow data access.
        bpy.ops.object.mode_set(mode='OBJECT');

        # Set the default return state to FINISHED
        result = {'FINISHED'};

        # Check that the currently selected object contains mesh data for exporting
        obj = context.active_object;
        if not obj:
            raise NameError("Cannot export: select 1 active object");

        if obj.type != 'MESH':
            raise NameError("Cannot export: active object is not a mesh");

        mesh = obj.data.copy();
        if not mesh:
            raise NameError("Cannot export: active object has no mesh data");

        mesh.calc_normals();

        # make sure we have tesselated faces
        if not mesh.tessfaces and mesh.polygons:
            mesh.calc_tessface();

        mesh.transform(obj.matrix_world);

        # faces and vertices are now in
        # mesh.tessfaces, mesh.tessfaces.vertices (len==3 oder 4)
        # obj.vertex_groups

        has_uv = bool(mesh.tessface_uv_textures)
        if has_uv:
            active_uv_layer = mesh.tessface_uv_textures.active
            if not active_uv_layer:
                has_uv = False
            else:
                active_uv_layer = active_uv_layer.data

        arm = None;
        arm_obj = obj.find_armature();
        if arm_obj and arm_obj.type == 'ARMATURE':
            arm = arm_obj.data;

        bones = None;
        if arm:
            bones = arm.bones.values();
            if not arm_obj.vertex_groups:
                print('Warning: Armature is not using vertex groups.')

        vg_names = {vgroup.index: vgroup.name for vgroup in obj.vertex_groups};
        vgroups = {vgroup.index: [] for vgroup in obj.vertex_groups};
        for v in mesh.vertices:
            for g in v.groups:
                vgroups[g.group] += [(v.index, g.weight)]


        ###################################################################
        # Open the file for writing
        file = open(self.filepath, 'wt');

        file.write('<mesh name="%s">\n' % obj.name)

        # faces
        for fidx, face in enumerate(mesh.tessfaces):
            file.write('\t<face>\n');
            if has_uv:
                uv = active_uv_layer[fidx]
                uv = uv.uv1, uv.uv2, uv.uv3, uv.uv4

            for vidx, verti in enumerate(face.vertices):
                vert = mesh.vertices[verti];
                file.write('\t\t<vert>\n')
                file.write('\t\t\t<pos x="%.6f" y="%.6f" z="%.6f" />\n' % tuple(vert.co))
                
                if has_uv:
                    file.write('\t\t\t<uv x="%.6f" y="%.6f" />\n' % tuple(uv[vidx]))
                
                # TODO normals are a bit off sometimes...
                file.write('\t\t\t<norm x="%.6f" y="%.6f" z="%.6f" />\n' % tuple(vert.normal))

                # bones
                vgiter = 0
                numbones = 0
                for vgi in vgroups:
                    vg = vgroups[vgi]
                    for (vgi,vgw) in vg:
                        #if vgi == vert.index:
                        if vgi == verti:
                            numbones = numbones + 1
                            if numbones > 4:
                                print('Warning: More than 4 bones applied on a vertex (%s)' % vg_names[vgiter])
                            else:
                                file.write('\t\t\t<bone name="%s" wgt="%.6f" />\n' % (vg_names[vgiter], vgw))
                    vgiter = vgiter + 1

                file.write('\t\t</vert>\n')

            file.write('\t</face>\n')

        # armature frames
        if bones:
            scene = bpy.context.scene
            curframe = scene.frame_current
            file.write('\t<armature frames="%d">\n' % (scene.frame_end - scene.frame_start + 1))
            
            for bone in bones:
                if not bone.parent:    # Start with the root bone(s) and recurse
                    writeBone(file,bone,arm_obj,True)
                    
            # Write keyframe marker names
            # (When animating, use markers in the timeline to set
            # keyframes at the start of each separate animation in the file
            timeline = {t.frame: t.name for t in scene.timeline_markers}
            for frame in timeline:
                file.write('\t\t<anim frame="%d" name="%s" />\n' % (frame, timeline[frame]))

            file.write('\t</armature>\n')
            scene.frame_set(curframe)
            
        # materials
        for mat in mesh.materials:
            if mat:
                for mtex in mat.texture_slots:
                    if mtex and mtex.texture and mtex.texture.image:
                        imgtype = 'Diffuse'
                        if mtex.texture.use_normal_map:          # Could get bunches of other information, just depends on what I need
                            imgtype = 'Normal'
                        file.write('\t<material type="%s" file="%s" />\n' % (imgtype, mtex.texture.image.filepath))

        file.write('</mesh>\n')

        # Close the file
        file.close();

        return result;

def writeBone(file, bone, arm_obj, recurse):
    scene = bpy.context.scene
    pose = arm_obj.pose
    file.write('\t\t<bonedef name="%s">\n' % bone.name)
    
    # Write each frame 
    for f in range(scene.frame_start, scene.frame_end + 1):
        scene.frame_set(f)
        
        posebone = pose.bones[bone.name]

        pose_matrix = arm_obj.matrix_world.inverted() * posebone.matrix_basis * arm_obj.matrix_world
        pose_matrix.transpose()
        file.write('\t\t\t<frame num="%d"\n' % f)
        
        file.write('\t\t\t\tm00="%.6f" m01="%.6f" m02="%.6f" m03="%.6f"\n' % tuple(pose_matrix[0]))
        file.write('\t\t\t\tm10="%.6f" m11="%.6f" m12="%.6f" m13="%.6f"\n' % tuple(pose_matrix[1]))
        file.write('\t\t\t\tm20="%.6f" m21="%.6f" m22="%.6f" m23="%.6f"\n' % tuple(pose_matrix[2]))
        file.write('\t\t\t\tm30="%.6f" m31="%.6f" m32="%.6f" m33="%.6f"\n' % tuple(pose_matrix[3]))
        # TODO position data seems to be missing?

        file.write('\t\t\t/>\n')

        # Skip bone weight for now, it doesn't seem to affect anything
    
    file.write('\t\t</bonedef>\n')

    if recurse and bone.children:
        for child in bone.children:
            writeBone(file,child,arm_obj,False)

def menu_func(self, context):
    self.layout.operator(EldritchXMLMeshExporter.bl_idname, text="Eldritch XML Mesh (.mesh)");

def register():
    bpy.utils.register_module(__name__);
    bpy.types.INFO_MT_file_export.append(menu_func);

def unregister():
    bpy.utils.unregister_module(__name__);
    bpy.types.INFO_MT_file_export.remove(menu_func);

if __name__ == "__main__":
    register()
